from typing import List, Union

import pandas as pd
from astropy import units
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time as ATime

from .base_logger import logger
from .fits import FitsArray


class Calculator:
    def __init__(self, fits_array: FitsArray) -> None:
        """
        Constructor method.
        Creates a Calculator Object.

        Parameters
        ----------
        fits_array: FitsArray
            A FitsArray.
        """
        logger.info(f"Creating an instance from {self.__class__.__name__}")

        self.fits_array = fits_array

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id: {id(self)}, fits_array: {self.fits_array})"

    def __repr__(self) -> str:
        return self.__str__()

    def jd(self, key: str, new_key: str = "JD", date_format: str = "isot", scale: str = "utc") -> None:
        """
        Inserts a header wth key of new_key and value of JD which calculated from key.

        Parameters
        ----------
        key: str
            The key where DATE (UTC) is stored.
        new_key: str, optional
            New key name for JD to be inserted. Default: JD.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.
        Returns
        -------
        None
            None.
        """
        logger.info(f"Calculating JD. Parameters: {key=}, {new_key=}, {date_format=}, {scale=}")
        times = self.fits_array.hselect(key)
        if len(times) == 0:
            return

        jds = ATime(times[key].to_numpy().tolist(), format=date_format, scale=scale).jd
        for fits, jd in zip(self.fits_array, jds):
            fits.hedit(new_key, str(jd))

    @classmethod
    def jd_c(cls, dates: Union[str, List[str]], date_format: str = "isot", scale: str = "utc") -> pd.DataFrame:
        """
        Returns JD of the given DATEs.

        Parameters
        ----------
        dates: str or List[str]
            DATE or List of DATEs (utc).
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        pd.DataFrame
            List of JDs.
        """
        logger.info(f"Calculating JD. Parameters: {dates=}, {date_format=}, {scale=}")
        jds = ATime(dates, format=date_format, scale=scale).jd
        return pd.DataFrame({"jd": jds})

    def hjd(self, key: str, position: SkyCoord, new_key: str = "HJD", date_format: str = "isot",
            scale: str = "utc") -> None:
        """
        Inserts a header wth key of new_key and value of HJD which calculated from key.

        Parameters
        ----------
        key: str
            The key where DATE (UTC) is stored.
        position: SkyCoord
            SkyCoord object of the Object.
        new_key: str, optional
            New key name for HJD to be inserted. Default: HJD.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        None
            None.
        """
        logger.info(f"Calculating JD. Parameters: {key=}, {position=}, {new_key=}, {date_format=}, {scale=}")
        times = self.fits_array.hselect(key)
        if len(times) == 0:
            return

        times = ATime(times[key].to_numpy().tolist(), format=date_format, scale=scale)
        ltt_helio = times.light_travel_time(position, 'heliocentric')
        times_heliocentre = times.utc + ltt_helio

        for fits, hjd in zip(self.fits_array, times_heliocentre):
            fits.hedit(new_key, str(hjd))

    @classmethod
    def hjd_c(cls, dates: Union[str, List[str]], position: SkyCoord, new_key: str = "HJD", date_format: str = "isot",
              scale: str = "utc") -> pd.DataFrame:
        """
        Inserts a header wth key of new_key and value of HJD which calculated from key.

        Parameters
        ----------
        dates: str or List[str]
            DATE or List of DATEs (utc).
        position: SkyCoord
            SkyCoord object of the Object.
        new_key: str, optional
            New key name for HJD to be inserted. Default: HJD.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        pd.DataFrame
            List of HJDs.
        """
        logger.info(f"Calculating JD. Parameters: {dates=}, {position=}, {new_key=}, {date_format=}, {scale=}")
        times = ATime(dates, format=date_format, scale=scale)
        ltt_helio = times.light_travel_time(position, 'heliocentric')
        times_heliocentre = times.utc + ltt_helio
        return pd.DataFrame({"hjd": times_heliocentre})

    def bjd(self, key: str, position: SkyCoord, new_key: str = "BJD", date_format: str = "isot",
            scale: str = "utc") -> None:
        """
        Inserts a header wth key of new_key and value of BJD which calculated from key.

        Parameters
        ----------
        key: str
            The key where DATE (utc) is stored.
        position: SkyCoord
            SkyCoord object of the Object.
        new_key: str, optional
            New key name for HJD to be inserted. Default: BJD.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        None
            None.
        """
        logger.info(f"Calculating JD. Parameters: {key=}, {position=}, {new_key=}, {date_format=}, {scale=}")
        times = self.fits_array.hselect(key)
        if len(times) == 0:
            return

        times = ATime(times[key].to_numpy().tolist(), format=date_format, scale=scale)
        ltt_helio = times.light_travel_time(position)
        times_heliocentre = times.utc + ltt_helio

        for fits, hjd in zip(self.fits_array, times_heliocentre):
            fits.hedit(new_key, str(hjd))

    @classmethod
    def bjd_c(cls, dates: Union[str, List[str]], position: SkyCoord, new_key: str = "BJD", date_format: str = "isot",
              scale: str = "utc") -> pd.DataFrame:
        """
        Inserts a header wth key of new_key and value of BJD which calculated from key.

        Parameters
        ----------
        dates: str or List[str]
            DATE or List of DATEs (utc).
        position: SkyCoord
            SkyCoord object of the Object.
        new_key: str, optional
            New key name for BJD to be inserted. Default: BJD.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        pd.DataFrame
            List of BJDs.
        """
        logger.info(f"Calculating JD. Parameters: {dates=}, {position=}, {new_key=}, {date_format=}, {scale=}")
        times = ATime(dates, format=date_format, scale=scale)
        ltt_helio = times.light_travel_time(position)
        times_heliocentre = times.utc + ltt_helio
        return pd.DataFrame({"hjd": times_heliocentre})

    def astropy_time(self, key: str, date_format: str = "isot", scale: str = "utc") -> ATime:
        """
        Returns a list of astropy.time.Time from given key in header.

        Parameters
        ----------
        key: str
            The key where DATE (utc) is stored.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        astropy.time.Time
            Time object.
        """
        logger.info(f"Converting to astropy.time.Time. Parameters: {key=}, {date_format=}, {scale=}")
        times = self.fits_array.hselect(key)
        if len(times) == 0:
            raise ValueError("Time not found")

        return ATime(times[key].to_numpy().flatten().tolist(), format=date_format, scale=scale)

    @classmethod
    def astropy_time_c(cls, dates: Union[str, List[str]], date_format: str = "isot", scale: str = "utc") -> ATime:
        """
        Returns a list of astropy.time.Time from given DATEs in header.

        Parameters
        ----------
        dates: str or List[str]
            The key where DATE (utc) is stored.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        astropy.time.Time
            Time object.
        """
        logger.info(f"Converting to astropy.time.Time. Parameters: {dates=}, {date_format=}, {scale=}")
        if len(dates) == 0:
            raise ValueError("Time not found")

        return ATime(dates, format=date_format, scale=scale)

    def sec_z(self, key: str, location: EarthLocation, position: SkyCoord, new_key: str = "ARIMASS",
              date_format: str = "isot", scale: str = "utc") -> None:
        """
        Inserts a header wth key of new_key and value of secz which calculated from key.

        Parameters
        ----------
        key: str
            The key where DATE (utc) is stored.
        location: EarthLocation
            EarthLocation of the Observation site.
        position: SkyCoord
            SkyCoord object of the Object.
        new_key: str, optional
            New key name for SECZ to be inserted. Default: ARIMASS.
        date_format: str, optional
            Time format of the DATE. Default: isot.
        scale: str, optional
            Scale of the DATEs. Default: utc.

        Returns
        -------
        None
            None.
        """
        logger.info(
            f"Calculating secz. Parameters: {key=}, {location=}, {position=}, {new_key=}, {date_format=}, {scale=}")
        times = self.astropy_time(key, date_format=date_format, scale=scale)

        frame = AltAz(obstime=times, location=location)
        obj_alt_az = position.transform_to(frame)
        obj_alt = obj_alt_az.secz
        seczs = obj_alt.value.tolist()
        for fits, secz in zip(self.fits_array, seczs):
            fits.hedit(new_key, str(secz))

    @classmethod
    def sec_z_c(cls, times: ATime, location: EarthLocation, position: SkyCoord) -> pd.DataFrame:
        """
        Returns secz which calculated from DATEs and given location and position.

        Parameters
        ----------
        times: astropy.time.Time
            List of dates.
        location: EarthLocation
            EarthLocation of the Observation site.
        position: SkyCoord
            SkyCoord object of the Object.

        Returns
        -------
        pd.DataFrame
            List of secz.
        """
        logger.info(f"Calculating secz. Parameters: {times=}, {location=}, {position=}")
        frame = AltAz(obstime=times, location=location)
        obj_alt_az = position.transform_to(frame)
        obj_alt = obj_alt_az.secz
        return pd.DataFrame({"secz": obj_alt.value.tolist()})


class Coordinates:
    @classmethod
    def location_from_name(cls, name: str) -> EarthLocation:
        """
        Returns an EarthLocation from a given name.

        Parameters
        ----------
        name: str
            Name of the site.

        Returns
        -------
        EarthLocation
            Location.
        """
        logger.info(f"Creating EarthLocation. Parameters: {name=}")
        return EarthLocation.of_site(name)

    @classmethod
    def location(cls, longitude: float, latitude: float, altitude: float = 0) -> EarthLocation:
        """
        Returns an EarthLocation from given longitude, latitude and altitude.

        Parameters
        ----------
        longitude: float
            longitude of the location.
        latitude: float
            latitude of the location.
        altitude: float
            altitude of the location.

        Returns
        -------
        EarthLocation
            Location.
        """
        logger.info(f"Creating EarthLocation. Parameters: {longitude=}, {latitude=}, {altitude=}")
        return EarthLocation(longitude * units.deg, latitude * units.deg, altitude * units.m)

    @classmethod
    def position_from_name(cls, name: str) -> SkyCoord:
        """
        Returns a SkyCoord from the given name.

        Parameters
        ----------
        name: str
            Name of the celestial object.

        Returns
        -------
        SkyCoord
            Position.
        """
        logger.info(f"Creating SkyCoord. Parameters: {name=}")
        return SkyCoord.from_name(name, frame="icrs")

    @classmethod
    def position(cls, ra: float, dec: float) -> SkyCoord:
        """
        Returns a SkyCoord from given ra and dec.

        Parameters
        ----------
        ra: float
            right ascension.
        dec: float
            declination.

        Returns
        -------
        SkyCoord
            Position.
        """
        logger.info(f"Creating SkyCoord. Parameters: {ra=}, {dec=}")
        return SkyCoord(ra=ra, dec=dec, unit=(units.hourangle, units.deg), frame="icrs")



from .calculator import Calculator, Coordinates
from .calibration import Calibration
from .combine import Combine
from .fits import Fits, FitsArray
from .photometry import APhot
from .utils import Fixer



from subprocess import PIPE

from pyraf import iraf

from .base_logger import logger
from .errors import ImageCountError
from .fits import Fits, FitsArray
from .utils import Check, Fixer


class Combine:
    def __init__(self, fits_array: FitsArray) -> None:
        """
        Constructor method
        Creates a Combine Object.

        Parameters
        ----------
        fits_array: FitsArray
            A FitsArray.
        """
        logger.info(f"Creating an instance from {self.__class__.__name__}")
        if len(fits_array) < 1:
            raise ImageCountError("There is no image to process")

        self.fits_array = fits_array
        iraf.noao(Stdout=PIPE)
        iraf.imred(Stdout=PIPE)
        iraf.ccdred(Stdout=PIPE)
        iraf.imutil(Stdout=PIPE)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id: {id(self)}, fits_array: {self.fits_array})"

    def __repr__(self) -> str:
        return self.__str__()

    def combine(self, operation: str, output: str = None, override: bool = False, reject: str = None) -> Fits:
        """
        Returns the combined Fits of FitsArray.

        Parameters
        ----------
        operation: str
            Type of operation.
        output: str, optional
            path of the new fits file.
        override: bool, optional
            Force (overwrite) the output if a file with the same name already exist.
        reject: str, optional
            Rejection method.

        Returns
        -------
        Fits
            Combined Fits of FitsArray.
        """
        logger.info(f"combine started. Parameters: {operation=}, {output=}, {override=}, {reject=}")
        Check.operation(operation)
        Check.rejection(reject)

        reject = Fixer.nonify(reject)

        if (not Check.is_none(reject)) and len(self.fits_array) < 3:
            logger.error("Not enough image found")
            raise ImageCountError("Not enough image found")

        output = Fixer.output(output, override=override, delete=True, prefix="irony_", suffix=".fits")

        iraf.noao.imred.ccdred.combine.unlearn()
        with self.fits_array.at_file() as at_file:
            iraf.noao.imred.ccdred.combine(f"'@{at_file}'", output=output, combine=operation, reject=reject, ccdtype="")

        return Fits.from_path(output)

    def zerocombine(self, operation: str, output: str = None, override: bool = False, reject: str = None) -> Fits:
        """
        Returns the zerocombine Fits of FitsArray.

        Parameters
        ----------
        operation: str
            Type of operation.
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            Force (overwrite) the output if a file with the same name already exist.
        reject: str, optional
            Rejection method.

        Returns
        -------
        Fits
            Combined Fits of FitsArray.
        """
        logger.info(f"zerocombine started. Parameters: {operation=}, {output=}, {override=}, {reject=}")
        Check.operation(operation)
        Check.rejection(reject)

        reject = Fixer.nonify(reject)

        if (not Check.is_none(reject)) and len(self.fits_array) < 3:
            logger.error("Not enough image found")
            raise ImageCountError("Not enough image found")

        output = Fixer.output(output, override=override, delete=True, prefix="irony_", suffix=".fits")

        iraf.noao.imred.ccdred.zerocombine.unlearn()
        with self.fits_array.at_file() as at_file:
            iraf.noao.imred.ccdred.zerocombine(f"'@{at_file}'", output=output, combine=operation, reject=reject,
                                               ccdtype="")

        return Fits.from_path(output)

    def darkcombine(self, operation: str, output: str = None, override: bool = False, reject: str = None,
                    scale: str = None) -> Fits:
        """
        Returns the darkcombine Fits of FitsArray.

        Parameters
        ----------
        operation: str
            Type of operation.
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            Force (overwrite) the output if a file with the same name already exist.
        reject: str, optional
            Rejection method.
        scale: str, optional
            Scaling method.

        Returns
        -------
        Fits
            Combined Fits of FitsArray.
        """
        logger.info(
            f"darkcombine started. Parameters: {operation=}, {output=}, {override=}, {reject=}, {scale=}"
        )
        Check.operation(operation)
        Check.rejection(reject)
        Check.scale(scale)

        reject = Fixer.nonify(reject)
        scale = Fixer.nonify(scale)

        if (not Check.is_none(reject)) and len(self.fits_array) < 3:
            logger.error("Not enough image found")
            raise ImageCountError("Not enough image found")

        output = Fixer.output(output, override=override, delete=True, prefix="irony_", suffix=".fits")

        iraf.noao.imred.ccdred.darkcombine.unlearn()
        with self.fits_array.at_file() as at_file:
            iraf.noao.imred.ccdred.darkcombine(f"'@{at_file}'", output=output, combine=operation, reject=reject,
                                               scale=scale, ccdtype="", process="no")

        return Fits.from_path(output)

    def flatcombine(self, operation: str, output: str = None, override: bool = False, reject: str = None,
                    scale: str = None) -> Fits:
        """
        Returns the flatcombine Fits of FitsArray.

        Parameters
        ----------
        operation: str
            Type of operation.
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            Force (overwrite) the output if a file with the same name already exist.
        reject: str, optional
            Rejection method.
        scale: str, optional
            Scaling method.

        Returns
        -------
        Fits
            Combined Fits of FitsArray.
        """
        logger.info(f"flatcombine started. Parameters: {operation=}, {output=}, {override=}, {reject=}, {scale=}")
        Check.operation(operation)
        Check.rejection(reject)
        Check.scale(scale)

        reject = Fixer.nonify(reject)
        scale = Fixer.nonify(scale)

        if (not Check.is_none(reject)) and len(self.fits_array) < 3:
            logger.error("Not enough image found")
            raise ImageCountError("Not enough image found")

        output = Fixer.output(output, override=override, delete=True, prefix="irony_", suffix=".fits")

        iraf.noao.imred.ccdred.flatcombine.unlearn()
        with self.fits_array.at_file() as at_file:
            iraf.noao.imred.ccdred.flatcombine(f"'@{at_file}'", output=output, combine=operation, reject=reject,
                                               scale=scale, ccdtype="", process="no")

        return Fits.from_path(output)

    def imsum(self, output: str = None, override: bool = False) -> Fits:
        """
        Returns the imsum Fits of FitsArray.

        Parameters
        ----------
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            Force (overwrite) the output if a file with the same name already exist.

        Returns
        -------
        Fits
            Sum of FitsArray.
        """
        logger.info(f"imsum started. Parameters:{output=}, {override=}")
        if len(self.fits_array) < 2:
            logger.error("Not enough image found")
            raise ImageCountError("Not enough image found")

        output = Fixer.output(output, override=override, delete=True, prefix="irony_", suffix=".fits")

        iraf.imutil.imsum.unlearn()
        with self.fits_array.at_file() as at_file:
            iraf.imutil.imsum(f"'@{at_file}'", output=output)
        return Fits.from_path(output)



import logging
import time

LOG_FORMAT = "[%(asctime)s, %(levelname)s],"
LOG_FORMAT += "[%(filename)s, %(funcName)s, %(lineno)s]: %(message)s"
logger = logging
logging.Formatter.converter = time.gmtime
logger.basicConfig(format=LOG_FORMAT, level=logging.ERROR)



from __future__ import annotations

import contextlib
import subprocess
import tempfile
from glob import glob
from pathlib import Path
from subprocess import PIPE
from typing import Dict, List, Union, Hashable

import astroalign
import matplotlib.animation as animation
import numpy as np
import pandas as pd
from astropy.io import fits as fts
from astropy.stats import sigma_clipped_stats
from astropy.visualization import ZScaleInterval
from matplotlib import pyplot as plt
from mpl_point_clicker import clicker
from photutils.detection import DAOStarFinder
from pyraf import iraf
from sep import Background, extract as sep_extract
from ccdproc import cosmicray_lacosmic

from .base_logger import logger
from .errors import AlignError, ImageCountError, NumberOfElementError
from .utils import Check, Fixer


class Fits:
    def __init__(self, path: Path):
        """
        Constructor method.
        Creates a Fits Object. The file_path must exist.

        Parameters
        ----------
        path: pathlib.Path
            Path object of the fits file.
        """
        self.NAME = self.__class__.__name__
        logger.info(f"Creating an instance from {self.NAME}")
        if not path.exists():
            raise FileNotFoundError("File does not exist")

        self.path = path

        iraf.noao(Stdout=PIPE)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(@: {id(self)}, file: {abs(self)})"

    def __repr__(self) -> str:
        return self.__str__()

    def __abs__(self) -> str:
        return str(self.path.absolute())

    def __sub__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "-")

    def __add__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "+")

    def __radd__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "+")

    def __mul__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "*")

    def __rmul__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "*")

    def __truediv__(self, other: Union[Fits, float, int]) -> Fits:
        return self.imarith(other, "/")

    @classmethod
    def from_path(cls, path: str) -> Fits:
        """
        Creates a Fits Object. The file_path must exist.

        Parameters
        ----------
        path: str
            Path of the fits file.

        Returns
        -------
        Fits
            Fits Object.
        """
        logger.info(f"Creating Fits from path. Parameters: {path=}")
        return Fits(Path(path))

    @property
    def file(self):
        return f"{{OBSCURED PATH}}/{str(self.path.stem)}{str(self.path.suffix)}"

    @property
    def imstat(self) -> dict:
        """
        Returns the npix, mean, stddev, min, max of the array as a dict. The default return of IRAF's imstatistics task.

        Returns
        -------
        dict
            Dictionary of statistics.
        """
        logger.info(f"imstat started. Parameters: None")

        iraf.imutil.imstatistics.unlearn()
        keys = ["npix", "mean", "stddev", "min", "max"]
        data = [each.split() for each in
                iraf.imutil.imstatistics(abs(self), fields="image,npix,mean,stddev,min,max", Stdout=PIPE)
                if not each.startswith("#") and not each.startswith("Error")]
        return {key: float(value) for key, value in zip(keys, data[0][1:])}

    @property
    def header(self) -> dict:
        """
        Returns the header of the fits file as a dict. The return of IRAF's imheader task with l+.

        Returns
        -------
        dict
            Dictionary of headers.
        """
        logger.info(f"Getting header. Parameters: None")

        header = fts.getheader(abs(self))
        return {i: header[i] for i in header if i}

    @property
    def data(self) -> np.ndarray:
        """
        Returns the header of the fits file as a np.array.

        Returns
        -------
        np.ndarray
            array of data.
        """
        logger.info(f"Getting data. Parameters: None")

        return fts.getdata(abs(self)).astype(float)

    def background(self, as_array: bool = False) -> Union[Background, np.ndarray]:
        """
        Returns the background object of the fits file.

        Parameters
        ----------
        as_array: bool
            If True returns a np.array of background. Otherwise, returns the object itself.

        Returns
        -------
        Union[Background, np.ndarray]
            Either a background object or a np.array.
        """
        logger.info(f"Getting background. Parameters: {as_array=}")

        if as_array:
            return Background(self.data).back()
        return Background(self.data)

    def cosmic_cleaner(self, output: str = None, override: bool = False, sigclip: float = 4.5, sigfrac: float = 0.3,
                       objlim: float = 5.0, gain: float = 1.0, readnoise: float = 6.5, satlevel: float = 65535.0,
                       pssl: float = 0.0, niter: int = 4, sepmed: bool = True, cleantype: str = 'meanmask',
                       fsmode: str = 'median', psfmodel: str = 'gauss', psffwhm: float = 2.5, psfsize: int = 7,
                       psfk: np.ndarray = None, psfbeta: float = 4.765, gain_apply: bool = True):
        """
        Returns consmic cleaned image.

        [1]: https://ccdproc.readthedocs.io/en/latest/api/ccdproc.cosmicray_lacosmic.html

        Parameters
        ----------
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            If True will overwrite the new_path if a file is already exists.
        sigclip: float, optional
            Laplacian-to-noise limit for cosmic ray detection. Lower values will flag more pixels as cosmic rays. Default: 4.5. see [1]
        sigfrac: float, optional
            Fractional detection limit for neighboring pixels. For cosmic ray neighbor pixels, a Laplacian-to-noise detection limit of sigfrac * sigclip will be used. Default: 0.3. see [1]
        objlim: float, optional
            Minimum contrast between Laplacian image and the fine structure image. Increase this value if cores of bright stars are flagged as cosmic rays. Default: 5.0. see [1]
        gain: float, optional
            Gain of the image (electrons / ADU). We always need to work in electrons for cosmic ray detection. Default: 1.0 see [1]
        readnoise: float, optional
            Read noise of the image (electrons). Used to generate the noise model of the image. Default: 6.5. see [1]
        satlevel: float, optional
            Saturation level of the image (electrons). This value is used to detect saturated stars and pixels at or above this level are added to the mask. Default: 65535.0. see [1]
        pssl: float, optional
            Previously subtracted sky level in ADU. We always need to work in electrons for cosmic ray detection, so we need to know the sky level that has been subtracted so we can add it back in. Default: 0.0. see [1]
        niter: float, optional
            Number of iterations of the LA Cosmic algorithm to perform. Default: 4. see [1]
        sepmed: int, optional
            Use the separable median filter instead of the full median filter. The separable median is not identical to the full median filter, but they are approximately the same, the separable median filter is significantly faster, and still detects cosmic rays well. Note, this is a performance feature, and not part of the original L.A. Cosmic. Default: True. see [1]
        cleantype: str, optional
            Set which clean algorithm is used: 1) "median": An unmasked 5x5 median filter. 2) "medmask": A masked 5x5 median filter. 3) "meanmask": A masked 5x5 mean filter. 4) "idw": A masked 5x5 inverse distance weighted interpolation. Default: "meanmask". see [1]
        fsmode: float, optional
            Method to build the fine structure image: 1) "median": Use the median filter in the standard LA Cosmic algorithm. 2) "convolve": Convolve the image with the psf kernel to calculate the fine structure image. Default: "median". see [1]
        psfmodel: str, optional
            Model to use to generate the psf kernel if fsmode == ‘convolve’ and psfk is None. The current choices are Gaussian and Moffat profiles: - "gauss" and "moffat" produce circular PSF kernels. - The "gaussx" and "gaussy" produce Gaussian kernels in the x and y directions respectively. Default: "gauss". see [1]
        psffwhm: float, optional
            Full Width Half Maximum of the PSF to use to generate the kernel. Default: 2.5. see [1]
        psfsize: int, optional
            Size of the kernel to calculate. Returned kernel will have size psfsize x psfsize. psfsize should be odd. Default: 7. see [1]
        psfk: np.ndarray (with float dtype), optional
            PSF kernel array to use for the fine structure image if fsmode == 'convolve'. If None and fsmode == 'convolve', we calculate the psf kernel using psfmodel. Default: None. see [1]
        psfbeta: float, optional
            Moffat beta parameter. Only used if fsmode=='convolve' and psfmodel=='moffat'. Default: 4.765.
        gain_apply: float, optional
            If True, return gain-corrected data, with correct units, otherwise do not gain-correct the data. Default is True to preserve backwards compatibility. see [1]

        Returns
        -------
        Fits
            Cleaned fits
        """
        logger.info(f"Starting Cosmic clean. Parameters: alot...")

        print(output)
        output = Fixer.output(output, override=override)
        print(output)

        newdata, _ = cosmicray_lacosmic(self.data, sigclip=sigclip, sigfrac=sigfrac, objlim=objlim, gain=gain,
                                        readnoise=readnoise, satlevel=satlevel, pssl=pssl, niter=niter, sepmed=sepmed,
                                        cleantype=cleantype, fsmode=fsmode, psfmodel=psfmodel, psffwhm=psffwhm,
                                        psfsize=psfsize, psfk=psfk, psfbeta=psfbeta, gain_apply=gain_apply)

        fts.writeto(output, newdata.value, header=fts.getheader(abs(self)))

        return Fits.from_path(output)

    def hedit(self, keys: Union[str, List[str]], values: Union[str, List[str]] = None, delete: bool = False,
              value_is_key: bool = False) -> None:
        """
        Edits header of the given file.

        Parameters
        ----------
        keys: str or List[str]
            Keys to be altered.
        values: str or List[str], optional
            Values to be added to set be set. Would be ignored if delete is True.
        delete: bool, optional
            Deletes the key from header if True.
        value_is_key: bool, optional
            Adds value of the key given in values if True. Would be ignored if delete is True.

        Returns
        -------
        None
            None.
        """
        logger.info(f"hedit started. Parameters: {keys=}, {values=}, {delete=}, {value_is_key=}, {keys=}")

        if delete:
            if isinstance(keys, str):
                keys = [keys]

            with fts.open(abs(self), "update") as hdu:
                for key in keys:
                    if key in hdu[0].header:
                        del hdu[0].header[key]
                    else:
                        logger.warning(f"{key} was not found in header. Skipping...")
        else:

            if not isinstance(values, type(keys)):
                logger.error(f"keys and values must both be strings or list of strings")
                raise ValueError("keys and values must both be strings or list of strings")

            if isinstance(keys, str):
                keys = [keys]

            if isinstance(values, str):
                values = [values]

            if len(keys) != len(values):
                logger.error(f"List of keys and values must be equal in length")
                raise ValueError("List of keys and values must be equal in length")

            with fts.open(abs(self), "update") as hdu:
                for key, value in zip(keys, values):
                    if value_is_key:
                        hdu[0].header[key] = hdu[0].header[value]
                    else:
                        hdu[0].header[key] = value

    def save_as(self, path: str, override: bool = False) -> Fits:
        """
        Saves the Fits file as new_path.

        Parameters
        ----------
        path: str
            New path to save the file.
        override: bool, optional
            If True will overwrite the new_path if a file is already exists.

        Returns
        -------
        Fits
            New fits object of saved fits file.
        """
        logger.info(f"saving as. Parameters: {path=}, {override=}")

        path = Fixer.output(path, override=override)

        iraf.imutil.imcopy.unlearn()
        iraf.imutil.imcopy(abs(self), path, verbose="no")

        return Fits.from_path(path)

    def imarith(self, other: Union[Fits, float, int], operand: str, output: str = None, override: bool = False) -> Fits:
        """
        Makes an arithmeic calculation on the file. The default behaviour of IRAF's imarith task.

        Parameters
        ----------
        other: Fits, float or int
            The value to be added to image array.
        operand: str
            An arithmetic operator. Either +, -, * or /.
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            If True will overwrite the new_path if a file is already exists.

        Returns
        -------
        Fits
            Fits object of resulting fits of the operation.
        """
        logger.info(f"imarith started. Parameters: {other=}, {operand=}, {output=}, {override=}")

        if not isinstance(other, (float, int, Fits)):
            logger.error(f"Please provide either a Fits Object or a numeric value")
            raise ValueError("Please provide either a Fits Object or a numeric value")

        Check.operand(operand)

        output = Fixer.output(output, override=override, delete=True, suffix=".fits", prefix="irony_")

        if isinstance(other, Fits):
            other = abs(other)

        iraf.imutil.imarith.unlearn()
        iraf.imutil.imarith(operand1=abs(self), op=operand, operand2=other, result=output)

        return Fits.from_path(output)

    def extract(self, detection_sigma: float = 5, min_area: float = 5):
        """
        Runs astroalign._find_sources to detect sources on the image.

        Parameters
        ----------
        detection_sigma: float
            `thresh = detection_sigma * bkg.globalrms`
        min_area: float
            Minimum area

        Returns
        -------
        pd.DataFrame
            List of sources found on the image.
        """
        bkg = self.background()
        thresh = detection_sigma * bkg.globalrms
        sources = sep_extract(self.data - bkg.back(), thresh, minarea=min_area)
        sources.sort(order="flux")
        if len(sources) < 0:
            raise NumberOfElementError("No source was found")

        return pd.DataFrame(
            sources,
            columns=["npix", "tnpix", "xmin", "xmax", "ymin", "ymax", "xcentroid", "ycentroid", "x2", "y2", "xy",
                     "errx2", "erry2", "errxy", "a", "b", "theta", "cxx", "cyy", "cxy", "cflux", "flux", "cpeak",
                     "peak", "xcpeak", "ycpeak", "xpeak", "ypeak", "flag"]
        )

    def daofind(self, sigma: float = 3, fwhm: float = 3, threshold: float = 5) -> pd.DataFrame:
        """
        Runs daofind to detect sources on the image.

        [1]: https://docs.astropy.org/en/stable/api/astropy.stats.sigma_clipped_stats.html

        [2]: https://photutils.readthedocs.io/en/stable/api/photutils.detection.DAOStarFinder.html

        Parameters
        ----------
        sigma: float, optional
            The number of standard deviations to use for both the lower and upper clipping limit. These limits are overridden by sigma_lower and sigma_upper, if input. The default is 3. [1]
        fwhm: float, optional
            The full-width half-maximum (FWHM) of the major axis of the Gaussian kernel in units of pixels. [2]
        threshold: float, optional
            The absolute image value above which to select sources. [2]

        Returns
        -------
        pd.DataFrame
            List of sources found on the image.
        """
        logger.info(
            f"daofind started. Parameters: {sigma=}, {fwhm=}, {threshold=}")

        mean, median, std = sigma_clipped_stats(self.data, sigma=sigma)
        daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
        sources = daofind(self.data - median)
        if sources is not None:
            return sources.to_pandas()
        return pd.DataFrame([],
                            columns=["id", "xcentroid", "ycentroid", "sharpness", "roundness1", "roundness2", "npix",
                                     "sky", "peak", "flux", "mag"])

    def align(self, other: Fits, output: str = None, max_control_points: int = 50, detection_sigma: float = 5,
              min_area: int = 5, override: bool = False) -> Fits:
        """
        Runs a Fits object of aligned Image.

        [1]: https://astroalign.quatrope.org/en/latest/api.html#astroalign.register

        Parameters
        ----------
        other: Fits
            The reference Image to be aligned as a Fits object.
        output: str, optional
            Path of the new fits file.
        max_control_points: int, optional
            The maximum number of control point-sources to find the transformation. [1]
        detection_sigma: int, optional
            Factor of background std-dev above which is considered a detection. [1]
        min_area: int, optional
            Minimum number of connected pixels to be considered a source. [1]
        override: bool, optional
            If True will overwrite the new_path if a file is already exists.

        Returns
        -------
        Fits
            Fits object of aligned image.
        """
        logger.info(
            f"align started. Parameters: {other=}, {output=}, {max_control_points=}, {detection_sigma=}, {min_area=}, {override=}")

        if not isinstance(other, Fits):
            logger.error("Other must be a Fits")
            raise ValueError("Other must be a Fits")

        output = Fixer.output(output, override=override)
        try:
            registered_image, footprint = astroalign.register(self.data, other.data,
                                                              max_control_points=max_control_points,
                                                              detection_sigma=detection_sigma, min_area=min_area)
            fts.writeto(output, registered_image, header=fts.getheader(abs(self)))
            return Fits.from_path(output)
        except ValueError:
            logger.error("Cannot align two images")
            raise AlignError("Cannot align two images")

    def show(self, points: pd.DataFrame = None, scale: bool = True) -> None:
        """
        Shows the Image using matplotlib.

        Parameters
        ----------
        points: pd.DataFrame, optional
            Draws points on image if a list is given.
        scale: bool, optional
            Scales the Image if True.

        Returns
        -------
        None
            None.
        """
        logger.info(f"showing image. Parameters: {points=}, {scale=}")

        if scale:
            zscale = ZScaleInterval()
        else:

            def zscale(x): return x

        plt.imshow(zscale(self.data), cmap="Greys_r")
        if points is not None:
            plt.scatter(points.xcentroid, points.ycentroid, s=10, c="red")
        plt.xticks([])
        plt.yticks([])
        plt.show()

    def solve_field(self, output: str = None, override: bool = False) -> None:
        # todo improve
        output = Fixer.output(output, override=override, delete=False)
        cmd = ["solve-field", str(self), "-o", output, "-p"]

        if override:
            cmd.append(override)

        process = subprocess.Popen(cmd)
        process.wait()

    def astrometry(self, api_key, solve_timeout=120):
        """
        Astrometry.net plate solving

        Parameters
        ----------
        api_key: str
            API Key
        solve_timeout: int
            Timeout

        Returns
        -------
        Fits
            Plate solved Fits object
        """
        pass
        # from astroquery.astrometry_net import AstrometryNet
        # ast = AstrometryNet()
        # ast.api_key = api_key
        #
        # try_again = True
        # submission_id = None
        #
        # while try_again:
        #     try:
        #         if not submission_id:
        #             wcs_header = ast.solve_from_image(abs(self),
        #                                               submission_id=submission_id)
        #         else:
        #             wcs_header = ast.monitor_submission(submission_id,
        #                                                 solve_timeout=solve_timeout)
        #     except TimeoutError as e:
        #         submission_id = e.args[1]
        #         print("="*82)
        #         print(e)
        #     else:
        #         try_again = False
        #
        # if wcs_header:
        #     print("asd")
        #     print(wcs_header)
        #     # with fts.open(abs(self), mode="update") as the_file:
        #     #     the_file[0].header.append(wcs_header)
        # else:
        #     print("dsa")
        #     print(wcs_header)

    def coordinate_picker(self, scale: bool = True) -> pd.DataFrame:
        """
        Shows the Image using matplotlib and returns a list of coordinates picked by user.

        Parameters
        ----------
        scale: bool, optional
            Scales the Image if True.

        Returns
        -------
        pd.DataFrame
            List of coordinates selected.
        """
        if scale:
            zscale = ZScaleInterval()
        else:

            def zscale(x):
                return x

        fig, ax = plt.subplots(constrained_layout=True)
        ax.imshow(zscale(self.data), cmap="Greys_r")
        klkr = clicker(ax, ["source"], markers=["o"])
        plt.show()
        if len(klkr.get_positions()["source"]) == 0:
            return pd.DataFrame([], columns=["xcentroid", "ycentroid"])

        return pd.DataFrame(
            klkr.get_positions()["source"], columns=[
                "xcentroid", "ycentroid"])


class FitsArray:
    def __init__(self, fits_list: List[Fits]) -> None:
        """
        Constructor method
        Creates a FitsArray Object. The length of  fits_lists must larger the 0.

        Parameters
        ----------
        fits_list: List[Fits]
            A list of Fits.
        """
        logger.info(f"Creating an instance from {self.__class__.__name__}")

        if not isinstance(fits_list, list):
            raise ValueError("fits_list must be a list of Fits")

        fits_list = [
            each
            for each in fits_list
            if isinstance(each, Fits)
        ]

        if len(fits_list) < 1:
            raise ImageCountError("No image was provided")

        self.fits_list = fits_list

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(@: {id(self)}, nof: {len(self)})"

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, key: int) -> Union[Fits, FitsArray]:
        element = self.fits_list[key]
        if isinstance(element, Fits):
            return element
        elif isinstance(element, List):
            return FitsArray(element)

        raise ValueError("Wrong slice")

    def __delitem__(self, key):
        del self.fits_list[key]

    def __len__(self) -> int:
        return len(self.fits_list)

    def __abs__(self) -> List[str]:
        return list(map(abs, self.fits_list))

    @classmethod
    def from_paths(cls, paths: List[str]) -> FitsArray:
        """
        Creates a FitsArray Object. The length of glob('file_path*.fit*') must be larger then 0.

        Parameters
        ----------
        paths: List[str]
            A list of strings of paths.

        Returns
        -------
        FitsArray
            FitsArray generated from list of paths as str.
        """
        logger.info(f"Creating FitsArray from from_paths. Parameters: {paths}")
        files = []
        for each in map(Path, paths):
            try:
                files.append(Fits(each))
            except FileNotFoundError:
                pass

        return FitsArray(files)

    @classmethod
    def from_pattern(cls, pattern: str) -> FitsArray:
        """
        Creates a FitsArray Object. The length of glob('file_path*.fit*') must be larger then 0.

        Parameters
        ----------
        pattern: str
            A pattern of a list of files.

        Returns
        -------
        FitsArray
            FitsArray generated from pattern of files.
        """
        logger.info(
            f"Creating FitsArray from from_paths. Parameters: {pattern}")

        return FitsArray.from_paths(glob(pattern))

    @contextlib.contextmanager
    def at_file(self) -> str:
        """
        Creates a text file with all fits file paths at each line. Useful for IRAF's @files.

        Returns
        -------
        str
            A context manager of a file containing file path of each file.
        """
        logger.info(f"Creating at_file. Parameters: None")

        with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".fls", prefix="irony_") as tmp:
            to_write = []
            for each in self.fits_list:
                to_write.append(abs(each))
            tmp.write("\n".join(to_write))
            tmp.flush()
            yield tmp.name

    @property
    def files(self):
        return [fits.file for fits in self.fits_list]

    @property
    def imstat(self) -> pd.DataFrame:
        """
        Returns the npix, mean, stddev, min, max of the array as a pd.DataFrame. The default return of
        IRAF's imstatistics task.

        Returns
        -------
        pd.DataFrame
            List of statistics of all files.
        """
        logger.info(f"imstat started. Parameters: None")

        iraf.imutil.imstatistics.unlearn()
        with self.at_file() as at_file:
            return (pd.DataFrame([each.split() for each in
                                  iraf.imutil.imstatistics(f"@{at_file}", fields="image,npix,mean,stddev,min,max",
                                                           Stdout=PIPE) if
                                  not each.startswith("#") and not each.startswith("Error")],
                                 columns=("image", "npix", "mean", "stddev", "min", "max")).set_index("image").replace(
                {np.nan: None}).astype(float))

    @property
    def header(self) -> pd.DataFrame:
        """
        Returns the header of the fits file(s) as a pd.DataFrame. The return of IRAF's imheader task with l+.

        Returns
        -------
        pd.DataFrame
            List of headers of all files.
        """
        logger.info(f"getting header. Parameters: None")

        headers = []
        for each in self:
            h = each.header
            h["image"] = str(abs(each))
            headers.append(h)

        return pd.DataFrame(headers).set_index("image").replace({np.nan: None})

    def hedit(self, keys: Union[str, List[str]], values: Union[str, List[str]] = None, delete: bool = False,
              value_is_key: bool = False) -> None:
        """
        Edits header of the given file.

        Parameters
        ----------
        keys: str or List[str]
            Keys to be altered.
        values: str or List[str], optional
            Values to be added to set be set. Would be ignored if delete is True.
        delete: bool, optional
            Deletes the key from header if True.
        value_is_key: bool, optional
            Adds value of the key given in values if True. Would be ignored if delete is True.

        Returns
        -------
        None
            None.
        """
        logger.info(f"hedit started. Parameters: {keys=}, {values=}, {delete=}, {value_is_key=}")

        if delete:
            if isinstance(keys, str):
                keys = [keys]
            for fits in self:
                with fts.open(abs(fits), "update") as hdu:
                    for key in keys:
                        if key in hdu[0].header:
                            del hdu[0].header[key]
                        else:
                            logger.warning(f"{key} was not found in header. Skipping for {abs(fits)}")
        else:
            if not isinstance(keys, type(values)):
                logger.error(f"keys and values must both be strings or list of strings")
                raise ValueError("keys and values must both be strings or list of strings")

            if isinstance(keys, str):
                keys = [keys]
                values = [values]

            if len(keys) != len(values):
                logger.error(f"List of keys and values must be equal in length")
                raise ValueError("List of keys and values must be equal in length")

            for fits in self:
                with fts.open(abs(fits), "update") as hdu:
                    for key, value in zip(keys, values):
                        if value_is_key:
                            hdu[0].header[key] = hdu[0].header[value]
                        else:
                            hdu[0].header[key] = value

    def hselect(self, fields: Union[str, List[str]]) -> pd.DataFrame:
        """
        Returns the header of the fits file(s) as a pd.DataFrame. The return of IRAF's imheader task with l+.

        Parameters
        ----------
        fields: str or List[str]
            Keys to be returned.

        Returns
        -------
        pd.DataFrame
            List of selected headers of all files.
        """
        logger.info(f"hselect started. Parameters: {fields=}")

        if isinstance(fields, str):
            fields = [fields]

        fields_to_use = []
        headers = self.header

        for field in fields:
            if field in headers.columns:
                fields_to_use.append(field)
            else:
                logger.warning(f"{field} was not found in header. Skipping...")

        if len(fields_to_use) < 1:
            return pd.DataFrame()
        return self.header[fields_to_use]

    def imarith(self, other: Union[FitsArray, Fits, float, int, List[float], List[int]], operand: str,
                output: str = None) -> FitsArray:
        """
        Makes an arithmeic calculation on the file(s). The default behaviour of IRAF's imarith task.

        Parameters
        ----------
        other: FitsArray, List[float], List[int], Fits, float or int
            The value to be added to image array.
        operand: str
            An arithmetic operator. Either +, -, * or /.
        output: str, optional
            Path of the new fits files.

        Returns
        -------
        FitsArray
            FitsArray object of resulting fits of the operation.
        """
        logger.info(f"imarith started. Parameters: {other=}, {operand=}, {output=}")

        if not isinstance(other, (float, int, FitsArray, Fits, List)):
            logger.error(f"Please provide either a FitsArray Object, Fits Object or a numeric value")
            raise ValueError("Please provide either a FitsArray Object, Fits Object or a numeric value")

        Check.operand(operand)

        iraf.imutil.imarith.unlearn()
        with self.at_file() as self_at:
            with Fixer.to_new_directory(output, self) as new_at:
                if isinstance(other, (Fits, float, int)):
                    if isinstance(other, Fits):
                        other = abs(other)
                    iraf.imutil.imarith(operand1=f"'@{self_at}'", op=f"'{operand}'", operand2=f"'{other}'",
                                        result=f"'@{new_at}'", verbose="no")
                else:
                    if isinstance(other, FitsArray):
                        with other.at_file() as other_at:
                            iraf.imutil.imarith(operand1=f"'@{self_at}'", op=f"'{operand}'", operand2=f"'@{other_at}'",
                                                result=f"'@{new_at}'", verbose="no")
                    else:
                        with Fixer.at_file_from_list(other) as other_at:
                            iraf.imutil.imarith(operand1=f"'@{self_at}'", op=f"'{operand}'", operand2=f"'@{other_at}'",
                                                result=f"'@{new_at}'", verbose="no")

                with open(new_at, "r") as new_files:
                    return FitsArray.from_paths(new_files.read().split())

    def align(self, other: Fits, output: str = None, max_control_points: int = 50, detection_sigma: float = 5,
              min_area: int = 5) -> FitsArray:
        """
        Runs a FitsArray object of aligned Image.

        [1]: https://astroalign.quatrope.org/en/latest/api.html#astroalign.register

        Parameters
        ----------
        other: Fits
            The reference Image to be aligned as a Fits.
        output: str, optional
            Path of the new fits files.
        max_control_points: int, optional
            The maximum number of control point-sources to find the transformation. [1]
        detection_sigma: int, optional
            Factor of background std-dev above which is considered a detection. [1]
        min_area: int, optional
            Minimum number of connected pixels to be considered a source. [1]

        Returns
        -------
        FitsArray
            FitsArray object of aligned images.
        """
        logger.info(
            f"align started. Parameters: {other=}, {output=}, {max_control_points=}, {detection_sigma=}, {min_area=}")

        if not isinstance(other, Fits):
            logger.error("Other must be a Fits")
            raise ValueError("Other must be a Fits")

        with Fixer.to_new_directory(output, self) as new_files:
            with open(new_files, "r") as f2r:
                aligned_files = []
                new_files = f2r.readlines()
                for fits, new_file in zip(self, new_files):
                    try:
                        new_fits = fits.align(other, new_file.strip(), max_control_points=max_control_points,
                                              detection_sigma=detection_sigma, min_area=min_area)
                        aligned_files.append(abs(new_fits))
                    except astroalign.MaxIterError:
                        pass
                    except AlignError:
                        pass
            if len(aligned_files) < 1:
                logger.error(f"None of the input images could be aligned")
                raise ImageCountError("None of the input images could be aligned")
            return FitsArray.from_paths(aligned_files)

    def cosmic_cleaner(self, output: str = None, override: bool = False, sigclip: float = 4.5, sigfrac: float = 0.3,
                       objlim: float = 5.0, gain: float = 1.0, readnoise: float = 6.5, satlevel: float = 65535.0,
                       pssl: float = 0.0, niter: int = 4, sepmed: bool = True, cleantype: str = 'meanmask',
                       fsmode: str = 'median', psfmodel: str = 'gauss', psffwhm: float = 2.5, psfsize: int = 7,
                       psfk: np.ndarray = None, psfbeta: float = 4.765, gain_apply: bool = True):
        """
        Returns consmic cleaned image array.

        [1]: https://ccdproc.readthedocs.io/en/latest/api/ccdproc.cosmicray_lacosmic.html

        Parameters
        ----------
        output: str, optional
            Path of the new fits file.
        override: bool, optional
            If True will overwrite the new_path if a file is already exists.
        sigclip: float, optional
            Laplacian-to-noise limit for cosmic ray detection. Lower values will flag more pixels as cosmic rays. Default: 4.5. see [1]
        sigfrac: float, optional
            Fractional detection limit for neighboring pixels. For cosmic ray neighbor pixels, a Laplacian-to-noise detection limit of sigfrac * sigclip will be used. Default: 0.3. see [1]
        objlim: float, optional
            Minimum contrast between Laplacian image and the fine structure image. Increase this value if cores of bright stars are flagged as cosmic rays. Default: 5.0. see [1]
        gain: float, optional
            Gain of the image (electrons / ADU). We always need to work in electrons for cosmic ray detection. Default: 1.0 see [1]
        readnoise: float, optional
            Read noise of the image (electrons). Used to generate the noise model of the image. Default: 6.5. see [1]
        satlevel: float, optional
            Saturation level of the image (electrons). This value is used to detect saturated stars and pixels at or above this level are added to the mask. Default: 65535.0. see [1]
        pssl: float, optional
            Previously subtracted sky level in ADU. We always need to work in electrons for cosmic ray detection, so we need to know the sky level that has been subtracted so we can add it back in. Default: 0.0. see [1]
        niter: int, optional
            Number of iterations of the LA Cosmic algorithm to perform. Default: 4. see [1]
        sepmed: float, optional
            Use the separable median filter instead of the full median filter. The separable median is not identical to the full median filter, but they are approximately the same, the separable median filter is significantly faster, and still detects cosmic rays well. Note, this is a performance feature, and not part of the original L.A. Cosmic. Default: True. see [1]
        cleantype: str, optional
            Set which clean algorithm is used: 1) "median": An unmasked 5x5 median filter. 2) "medmask": A masked 5x5 median filter. 3) "meanmask": A masked 5x5 mean filter. 4) "idw": A masked 5x5 inverse distance weighted interpolation. Default: "meanmask". see [1]
        fsmode: float, optional
            Method to build the fine structure image: 1) "median": Use the median filter in the standard LA Cosmic algorithm. 2) "convolve": Convolve the image with the psf kernel to calculate the fine structure image. Default: "median". see [1]
        psfmodel: str, optional
            Model to use to generate the psf kernel if fsmode == ‘convolve’ and psfk is None. The current choices are Gaussian and Moffat profiles: - "gauss" and "moffat" produce circular PSF kernels. - The "gaussx" and "gaussy" produce Gaussian kernels in the x and y directions respectively. Default: "gauss". see [1]
        psffwhm: float, optional
            Full Width Half Maximum of the PSF to use to generate the kernel. Default: 2.5. see [1]
        psfsize: int, optional
            Size of the kernel to calculate. Returned kernel will have size psfsize x psfsize. psfsize should be odd. Default: 7. see [1]
        psfk: np.ndarray (with float dtype), optional
            PSF kernel array to use for the fine structure image if fsmode == 'convolve'. If None and fsmode == 'convolve', we calculate the psf kernel using psfmodel. Default: None. see [1]
        psfbeta: float, optional
            Moffat beta parameter. Only used if fsmode=='convolve' and psfmodel=='moffat'. Default: 4.765.
        gain_apply: int, optional
            If True, return gain-corrected data, with correct units, otherwise do not gain-correct the data. Default is True to preserve backwards compatibility. see [1]

        Returns
        -------
            FitsArray
            Cleaned FitsArray
        """
        logger.info(f"Starting Cosmic clean. Parameters: alot...")

        with Fixer.to_new_directory(output, self) as new_files:
            with open(new_files, "r") as f2r:
                cleaned_files = []
                new_files = f2r.read().split()
                for fits, new_file in zip(self, new_files):
                    try:

                        new_fits = fits.cosmic_cleaner(output=new_file, override=override, sigclip=sigclip,
                                                       sigfrac=sigfrac, objlim=objlim, gain=gain, readnoise=readnoise,
                                                       satlevel=satlevel, pssl=pssl, niter=niter, sepmed=sepmed,
                                                       cleantype=cleantype, fsmode=fsmode, psfmodel=psfmodel,
                                                       psffwhm=psffwhm, psfsize=psfsize, psfk=psfk, psfbeta=psfbeta,
                                                       gain_apply=gain_apply)
                        cleaned_files.append(abs(new_fits))
                    except astroalign.MaxIterError:
                        pass
                    except AlignError:
                        pass
            if len(cleaned_files) < 1:
                logger.error(f"None of the input images could be cleaned")
                raise ImageCountError("None of the input images could be cleaned")
            return FitsArray.from_paths(cleaned_files)

    def show(self, scale: bool = True, interval: float = 1):
        """
        Animates the Images using matplotlib.

        Parameters
        ----------
        scale: bool, optional
            Scales the Image if True.
        interval: float, optional
            Interval of the animation. The smaller the value the faster the animation.

        Returns
        -------
        None
            None.
        """
        logger.info(f"animating images. Parameters: {scale=}, {interval=}")

        fig = plt.figure()

        if scale:
            zscale = ZScaleInterval()
        else:

            def zscale(x):
                return x

        im = plt.imshow(zscale(self[0].data), cmap="Greys_r", animated=True)
        plt.xticks([])
        plt.yticks([])

        def updatefig(args):
            im.set_array(zscale(self[args % len(self)].data))
            return im,

        _ = animation.FuncAnimation(fig, updatefig, interval=interval, blit=True)
        plt.show()

    def groupby(self, groups: Union[str, List[str]]) -> Dict[Hashable, FitsArray]:
        """
        Groups FitsArray by given key in header. Returns a dict with tuple of keys as key and FitsArray as value.
        
        Parameters
        ----------
        groups: str or List[str]
            Key(s).

        Returns
        -------
        dict
            A dictionary of grouped images.
        """
        logger.info(f"groupby started. Parameters: {groups=}")

        if isinstance(groups, str):
            groups = [groups]

        if len(groups) < 1:
            return dict()

        headers = self.header
        for group in groups:
            if group not in headers.columns:
                headers[group] = "N/A"

        grouped = {}
        for keys, df in headers.fillna("N/A").groupby(groups, dropna=False):
            grouped[keys] = FitsArray.from_paths(df.index.tolist())

        return grouped

    def save_as(self, output: str) -> FitsArray:
        """
        Saves the FitsArray files to output.

        Parameters
        ----------
        output: str
            New directory to save files.

        Returns
        -------
        FitsArray
            New FitsArray object of saved fits files.
        """
        logger.info(f"saving as. Parameters: {output=}")
        with self.at_file() as self_at:
            with Fixer.to_new_directory(output, self) as new_at:
                iraf.imutil.imcopy.unlearn()
                iraf.imutil.imcopy(f"'@{self_at}'", f"'@{new_at}'", verbose="no")

                with open(new_at, "r") as new_files:
                    return FitsArray.from_paths(new_files.read().split())



import pytest

from irony import Coordinates

DELTA = 10**-6

def test_location_from_name():
    site = Coordinates.location_from_name("TUG")
    assert site.lat.deg == pytest.approx(36.824167)
    assert site.lon.deg == pytest.approx(30.335556)

def test_location():
    site = Coordinates.location(45, 45, 2500)
    assert site.lat.deg == pytest.approx(45)
    assert site.lon.deg == pytest.approx(45)
    assert site.height.value == pytest.approx(2500)


def test_position_from_name():
    xy_leo = Coordinates.position_from_name("XY Leo")
    assert xy_leo.ra.hourangle == pytest.approx(10.027894)
    assert xy_leo.dec.deg == pytest.approx(17.409056)


def test_position():
    obj = Coordinates.position(10, 10)
    assert obj.ra.hourangle == pytest.approx(10)
    assert obj.dec.deg == pytest.approx(10)



import pytest

from irony import Calculator, Coordinates, FitsArray

FILES = "test/files/test*.fits"


def test_jd_c():
    fa = FitsArray.from_pattern(FILES)
    dates = fa.hselect("DATE-OBS").to_numpy().flatten().tolist()
    jds = Calculator.jd_c(dates).to_numpy().flatten().tolist()
    assert jds == pytest.approx([2456865.423257292, 2456865.421439815])


def test_sec_z_c():
    site = Coordinates.location(45, 45, 2000)
    v523_cas = Coordinates.position_from_name("v523 Cas")
    fa = FitsArray.from_pattern(FILES)
    dates = fa.hselect("DATE-OBS").to_numpy().flatten().tolist()
    secz = Calculator.sec_z_c(
        dates, site, v523_cas
    ).to_numpy().flatten().tolist()
    assert secz == pytest.approx([1.1844316720852472, 1.1899036959879108])



import pytest

from pathlib import Path
import numpy as np
from astropy.io import fits as afits
from astropy.stats import sigma_clipped_stats
from pandas.testing import assert_frame_equal
from photutils.detection import DAOStarFinder
from sep import Background

from irony import Fits

FILE = "test/files/test1.fits"


def test_abs():
    fits = Fits.from_path(FILE)
    assert abs(fits) == str(Path(FILE).absolute())


def test_doesnotexist():
    with pytest.raises(FileNotFoundError):
        _ = Fits.from_path("none_existing_path")

def test_imstat():
    fits = Fits.from_path(FILE)
    data = afits.getdata(FILE)

    assert fits.imstat == {
            'npix': data.size,
            'mean': np.round(np.mean(data), 1),
            'stddev': np.round(np.std(data), 1),
            'min': np.round(np.min(data), 1),
            'max': np.round(np.max(data), 1)
        }


def test_header():
    fits = Fits.from_path(FILE)

    h = afits.getheader(FILE)
    h_as_dict = {
        each: h[each]
        for each in h
        if each
    }

    assert fits.header == h_as_dict


def test_data():
    fits = Fits.from_path(FILE)
    np.testing.assert_equal(
        fits.data, afits.getdata(FILE)
    )


def test_background():
    fits = Fits.from_path(FILE)
    bkg = Background(afits.getdata(FILE).astype(float))

    assert isinstance(fits.background(), Background)

    np.testing.assert_equal(
        fits.background(as_array=True), bkg.back()
    )

def test_hedit():
    fits = Fits.from_path(FILE)

    h = afits.getheader(FILE)
    h_as_dict = {
        each: h[each]
        for each in h
        if each
    }

    assert fits.header == h_as_dict

    fits.hedit("IRON", "TEST")

    h = afits.getheader(FILE)
    h_as_dict = {
        each: h[each]
        for each in h
        if each
    }

    assert fits.header == h_as_dict
    assert "IRON" in fits.header

    fits.hedit("IRON", delete=True)

    h = afits.getheader(FILE)
    h_as_dict = {
        each: h[each]
        for each in h
        if each
    }

    assert fits.header == h_as_dict
    assert "IRON" not in fits.header

    fits.hedit("IRON", "DATE-OBS", value_is_key=True)

    h = afits.getheader(FILE)
    h_as_dict = {
        each: h[each]
        for each in h
        if each
    }

    assert fits.header == h_as_dict
    assert "IRON" in fits.header
    assert fits.header["IRON"] == fits.header["DATE-OBS"]

    fits.hedit("IRON", delete=True)
    assert "IRON" not in fits.header


def test_save_as():
    if Path("test/copy.fit").exists():
        Path("test/copy.fit").unlink()

    fits = Fits.from_path(FILE)
    fits.save_as("test/copy.fit")

    assert Path("test/copy.fit").exists()

    test_fits = Fits.from_path("test/copy.fit")

    np.testing.assert_equal(
        fits.data, test_fits.data
    )

    old_header = fits.header
    new_header = test_fits.header

    with pytest.raises(FileExistsError):
        fits.save_as("test/copy.fit")

    Path("test/copy.fit").unlink()


def test_imarith():
    fits = Fits.from_path(FILE)

    new_fits = fits.imarith(10, "*")
    np.testing.assert_equal(
        new_fits.data, fits.data * 10
    )

    new_fits = fits.imarith(2, "/")
    np.testing.assert_equal(
        new_fits.data, fits.data / 2
    )

    new_fits = fits.imarith(fits, "+")
    np.testing.assert_equal(
        new_fits.data, fits.data * 2
    )

    new_fits = fits.imarith(fits, "-")
    np.testing.assert_equal(
        new_fits.data, fits.data * 0
    )

    with pytest.raises(ValueError):
        _ = fits.imarith("Not supported value", "-")


def test_daofind():
    fits = Fits.from_path(FILE)
    sigma = 3
    threshold = 5
    fwhm = 3

    data = afits.getdata(FILE)
    mean, median, std = sigma_clipped_stats(data, sigma=sigma)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
    sources = daofind(data - median)

    assert_frame_equal(
        sources.to_pandas(),
        fits.daofind(sigma=sigma, fwhm=fwhm, threshold=threshold)
    )




from __future__ import annotations
import contextlib
import shutil
import tempfile
from glob import glob
from pathlib import Path, PurePath
from typing import List
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .fits import FitsArray

import pandas as pd

from .base_logger import logger
from .errors import (EmissionValueError, NoiseValueError, OperandValueError,
                     OperationValueError, RejectionValueError, ScaleValueError)


class Fixer:
    @classmethod
    def fitsify(cls, path: str) -> str:
        """
        Makes sure a file name ends either with fit or fits

        Parameters
        ----------
        path: str
            Path of the file

        Returns
        -------
        str
            New path
        """
        logger.info(f"fitsify started. Parameters: {path=}")

        if not (path.endswith("fit") or path.endswith("fits")):
            return f"{path}.fits"

        return path

    @classmethod
    def nonify(cls, value: str) -> str:
        """
        Converts None to "none"

        Parameters
        ----------
        value: str
            Any value

        Returns
        -------
        str
            new value
        """
        logger.info(f"nonify started. Parameters: {value=}")

        if value is None:
            return "none"

        return value

    @classmethod
    def output(cls, value: str, override: bool = False, delete: bool = True, prefix: str = "irony_", suffix: str = ".fits") -> str:
        """

        Parameters
        ----------
        value: str
        override: bool
        delete: bool
        prefix: str
        suffix: str

        Returns
        -------
        str
        """
        logger.info(f"output started. Parameters: {value=}, {override=}, {delete=}, {prefix=}, {suffix=}")

        if value is None:
            value = tempfile.NamedTemporaryFile(delete=delete, prefix=prefix, suffix=suffix).name

        value = cls.fitsify(value)

        if Path(value).exists():
            if override:
                Path(value).unlink()
            else:
                raise FileExistsError("File already exist")
        return value

    @classmethod
    @contextlib.contextmanager
    def to_new_directory(cls, output: str, fits_array: FitsArray) -> str:
        """

        Parameters
        ----------
        output: str
        fits_array: FitsArray

        Returns
        -------
        ste
        """
        logger.info(f"to_new_directory started. Parameters: {output=}, {fits_array=}")

        if output is None or not Path(output).is_dir():
            output = tempfile.mkdtemp(prefix="irony_")

        with tempfile.NamedTemporaryFile(delete=True, prefix="irony_", suffix=".fls", mode="w") as new_files_file:
            to_write = []
            for each_file in fits_array:
                f = each_file.path
                to_write.append(str(PurePath(output, f.name)))
            new_files_file.write("\n".join(to_write))
            new_files_file.flush()

            yield new_files_file.name

    @classmethod
    @contextlib.contextmanager
    def at_file_from_list(cls, data) -> str:
        """

        Parameters
        ----------
        data: list

        Returns
        -------
        str
        """
        logger.info(f"at_file_from_list started. Parameters: {data=}")

        with tempfile.NamedTemporaryFile(delete=True, prefix="irony_", suffix=".fls", mode="w") as new_files_file:
            new_files_file.write("\n".join(map(str, data)))
            new_files_file.flush()

            yield new_files_file.name

    @classmethod
    def yesnoify(cls, value: str) -> str:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        str
        """
        logger.info(f"yesnoify started. Parameters: {value=}")

        return "yes" if value else "no"

    @classmethod
    @contextlib.contextmanager
    def iraf_coords(cls, points: pd.DataFrame) -> str:
        """

        Parameters
        ----------
        points: pd.DataFrame

        Returns
        -------
        str
        """
        logger.info(f"iraf_coords started. Parameters: {points=}")
        with tempfile.NamedTemporaryFile(delete=True, prefix="irony_", suffix=".coo", mode="w") as new_files_file:
            # file_name = tempfile.NamedTemporaryFile(delete=False, prefix="irony_", suffix=".coo").name
            points[["xcentroid", "ycentroid"]].to_csv(new_files_file.name, sep=" ", header=False, index=False)
            new_files_file.flush()
            yield new_files_file.name

    @classmethod
    def list_to_source(cls, sources: List[List[float]]) -> pd.DataFrame:
        """

        Parameters
        ----------
        sources: List[List[float]]

        Returns
        -------
        pd.DataFrame
        """
        logger.info(f"list_to_source started. Parameters: {sources=}")

        return pd.DataFrame(sources, columns=["xcentroid", "ycentroid"])

    @classmethod
    def lists_to_source(cls, xs: List[float], ys: List[float]) -> pd.DataFrame:
        """

        Parameters
        ----------
        xs: List[float]
        ys: List[float]

        Returns
        -------
        pd.DataFrame
        """
        return pd.DataFrame({"xcentroid": xs, "ycentroid": ys})

    @classmethod
    def tmp_cleaner(cls):
        """
        Cleans tem folder
        Returns
        -------

        """
        for path in glob("/tmp/irony*"):
            if Path(path).is_file():
                Path(path).unlink()
            else:
                shutil.rmtree(path)


class Check:
    @classmethod
    def emision(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"emision checking. Parameters: {value=}")

        if not value.lower() in ["yes", "no"]:
            raise EmissionValueError("Emision value can only be one of: yes|no")

    @classmethod
    def noise(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"noise checking. Parameters: {value=}")

        if not value.lower() in ["poisson", "constant"]:
            raise NoiseValueError("Noise value can only be one of: poisson|constant")

    @classmethod
    def operation(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"operation checking. Parameters: {value=}")

        if value not in ["average", "median"]:
            raise OperationValueError("Operation value can only be one of: average|median")

    @classmethod
    def rejection(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"rejection checking. Parameters: {value=}")

        if value not in [ "none", "minmax", "ccdclip", "crreject", "sigclip", "avsigclip", "pclip", None]:
            raise RejectionValueError("Rejection value can only be one of: none|minmax|ccdclip|crreject|sigclip|avsigclip|pclip")

    @classmethod
    def operand(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"operand checking. Parameters: {value=}")

        if value not in ["+", "-", "*", "/"]:
            raise OperandValueError("Operand value can only be one of: +|-|*|/")

    @classmethod
    def scale(cls, value: str) -> None:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        None
            None.
        """
        logger.info(f"scale checking. Parameters: {value=}")

        if value not in ["none", "mode", "median", "mean", "exposure", None]:
            raise ScaleValueError("Scale value can only be one of: none|mode|median|mean|exposure")

    @classmethod
    def is_none(cls, value: str) -> str:
        """

        Parameters
        ----------
        value: str

        Returns
        -------
        str
        """
        logger.info(f"is_none checking. Parameters: {value=}")

        return value is None or value.lower() == "none"



from __future__ import annotations

import math
from subprocess import PIPE
from typing import Tuple, Union

import numpy as np
import pandas as pd
from photutils.aperture import (CircularAnnulus, CircularAperture, aperture_photometry)
from photutils.utils import calc_total_error
from pyraf import iraf
from sep import sum_circle

from .base_logger import logger
from .errors import NumberOfElementError
from .fits import FitsArray
from .utils import Fixer


class APhot:
    def __init__(self, fits_array: FitsArray) -> None:
        """
        Constructor method.
        Creates an Aperture Photometry Object.

        Parameters
        ----------
        fits_array: FitsArray
            A FitsArray.
        """
        logger.info("Creating an instance from APhot")
        self.fits_array = fits_array
        self.ZMag = 25

        iraf.digiphot(Stdout=PIPE)
        iraf.digiphot.apphot(Stdout=PIPE)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(id: {id(self)}, fits_array: {self.fits_array})"

    def __repr__(self) -> str:
        return self.__str__()

    def __flux2mag(self, flux: float, flux_error: float, exptime: float) -> Tuple[float, float]:
        logger.info(f"Converting to mag from flux. Parameters: {flux=}, {flux_error=}, {exptime=}")

        if exptime == 0:
            mag = self.ZMag + -2.5 * math.log10(flux)
        else:
            mag = self.ZMag + -2.5 * math.log10(flux) + 2.5 * math.log10(exptime)

        if flux_error <= 0:
            merr = 0
        else:
            merr = math.sqrt(flux / flux_error)

        if math.isinf(merr):
            merr = 0

        return mag, merr

    def __extract(self, keys: Union[str, list[str]]) -> pd.DataFrame:
        logger.info(f"Extracting header from FitsArray. Parameters: {keys=}")
        headers = self.fits_array.hselect(keys)
        return headers

    def photutils(self, points: pd.DataFrame, radius: int, radius_out: int = None,
                  extract: Union[str, list[str]] = None) -> pd.DataFrame:
        """
        Does photometry of given FitsArray using photutils and returns a pd.DataFrame.

        Parameters
        ----------
        points: pd.DataFrame
            A dataframe with x (xcentroid) and y (ycentroid) coordinates of sources for photometry.
        radius: float
            Aperture value.
        radius_out: float, optional
            Radius for sky measurements.
        extract: str o List[str], optional
            Headers to be extracted from fits files during photometry.

        Returns
        -------
        pd.DataFrame
            Photometric result.
        """
        logger.info(f"Photutils photometry. Parameters: {points=}, {radius=}, {radius_out=}, {extract=}")
        if len(points) < 1:
            logger.error("No coordinates were found")
            raise NumberOfElementError("No coordinates were found")

        table = []
        if radius_out is None:
            aperture = CircularAperture(points[["xcentroid", "ycentroid"]].to_numpy().tolist(), r=radius)
        else:
            aperture = CircularAnnulus(points[["xcentroid", "ycentroid"]].to_numpy().tolist(), r_in=radius,
                                       r_out=radius_out)
        for fits in self.fits_array:
            clean_d = fits.data - fits.background().rms()
            error = calc_total_error(fits.data, fits.background(as_array=True), fits.header["EXPTIME"])
            phot_table = aperture_photometry(fits.data, aperture, error=error)
            for line in phot_table:
                value = clean_d[int(line["xcenter"].value)][int(line["ycenter"].value)]
                snr = np.nan if value < 0 else math.sqrt(value)
                table.append([abs(fits), line["xcenter"].value, line["ycenter"].value,
                              *self.__flux2mag(line["aperture_sum"], line["aperture_sum_err"], fits.header["EXPTIME"]),
                              line["aperture_sum"], line["aperture_sum_err"], snr])

        phot_data = pd.DataFrame(table, columns=["image", "xcentroid", "ycentroid", "mag", "merr", "flux", "ferr",
                                                 "SNR"]).set_index("image")

        phot_data = phot_data.astype(float)

        if extract is not None:
            extracted_headers = self.__extract(extract)
            if len(extracted_headers) != 0:
                return pd.merge(phot_data, extracted_headers, left_index=True, right_index=True)

        return phot_data

    def sep(self, points: pd.DataFrame, radius: int, extract: list[str] = None) -> pd.DataFrame:
        """
        Does photometry of given FitsArray using sep and returns a pd.DataFrame.

        Parameters
        ----------
        points: pd.DataFrame
            A dataframe with x (xcentroid) and y (ycentroid) coordinates of sources for photometry.
        radius: float
            Aperture value.
        extract: str o List[str], optional
            Headers to be extracted from fits files during photometry.

        Returns
        -------
        pd.DataFrame
            Photometric result.
        """
        logger.info(f"sep photometry. Parameters: {points=}, {radius=}, {extract=}")
        if len(points) < 1:
            logger.error("No coordinates were found")
            raise NumberOfElementError("No coordinates were found")

        table = []
        for fits in self.fits_array:
            clean_d = fits.data - fits.background().rms()
            fluxes, ferrs, flag = sum_circle(fits.data, points["xcentroid"], points["ycentroid"], radius)
            for x, y, flux, ferr in zip(points["xcentroid"], points["ycentroid"], fluxes, ferrs):
                value = clean_d[int(x)][int(y)]
                snr = np.nan if value < 0 else math.sqrt(value)
                table.append([abs(fits), x, y, *self.__flux2mag(flux, ferr, fits.header["EXPTIME"]), flux, ferr, snr])

        phot_data = pd.DataFrame(table, columns=["image", "xcentroid", "ycentroid", "mag", "merr", "flux", "ferr",
                                                 "SNR"]).set_index("image")

        phot_data = phot_data.astype(float)

        if extract is not None:
            extracted_headers = self.__extract(extract)
            if len(extracted_headers) != 0:
                return pd.merge(phot_data, extracted_headers, left_index=True, right_index=True)

        return phot_data

    def iraf(self, points: pd.DataFrame, aperture: float, annulus: float, dannulu: float,
             extract: list[str] = None) -> pd.DataFrame:
        """
        Does photometry of given FitsArray using iraf and returns a pd.DataFrame.

        Parameters
        ----------
        points: pd.DataFrame
            A dataframe with x (xcentroid) and y (ycentroid) coordinates of sources for photometry.
        aperture: float
            Aperture value.
        annulus: float
            Annulus for sky measurements.
        dannulu: float
            Dannulu for sky measurements.
        extract: str o List[str], optional
            Headers to be extracted from fits files during photometry.

        Returns
        -------
        pd.DataFrame
            Photometric result.
        """
        logger.info("iraf photometry")
        if len(points) < 1:
            logger.error("No coordinates were found")
            raise NumberOfElementError("No coordinates were found")

        iraf.digiphot.apphot.datapars.unlearn()
        iraf.digiphot.apphot.centerpars.unlearn()
        iraf.digiphot.apphot.fitskypars.unlearn()
        iraf.digiphot.apphot.photpars.unlearn()
        iraf.digiphot.apphot.phot.unlearn()

        iraf.digiphot.apphot.datapars.gain = ""

        iraf.digiphot.apphot.centerpars.cbox = aperture / 2

        iraf.digiphot.apphot.fitskypars.salgori = "centroid"
        iraf.digiphot.apphot.fitskypars.annulus = annulus
        iraf.digiphot.apphot.fitskypars.dannulu = dannulu

        iraf.digiphot.apphot.photpars.weighting = "constant"
        iraf.digiphot.apphot.photpars.aperture = aperture
        iraf.digiphot.apphot.photpars.zmag = self.ZMag
        iraf.digiphot.apphot.photpars.mkapert = "no"
        with self.fits_array.at_file() as at_file:
            with Fixer.to_new_directory(None, self.fits_array) as output_files:
                with Fixer.iraf_coords(points[["xcentroid", "ycentroid"]]) as coo_file:
                    iraf.digiphot.apphot.phot(f"'@{at_file}'", coords=f"{coo_file}", output=f"'@{output_files}'",
                                              interac="no", verify="no")
                    res = iraf.txdump(f"'@{output_files}'", "id,mag,merr,flux,stdev", "yes", Stdout=PIPE)
                    res = pd.DataFrame([each.split() for each in res], columns=["id", "mag", "merr", "flux", "stdev"])

                    result = []

                    for each in res.groupby("id"):
                        coords = points.iloc[int(each[0]) - 1][["xcentroid", "ycentroid"]].tolist()
                        for i, fits in zip(range(len(each[1])), self.fits_array):
                            clean_d = fits.data - fits.background().rms()
                            phot = each[1].iloc[i]
                            value = clean_d[int(coords[0])][int(coords[0])]
                            snr = np.nan if value < 0 else math.sqrt(value)
                            result.append(
                                [abs(fits), coords[0], coords[1], phot.mag, phot.merr, phot.flux, phot.stdev, snr])

                    phot_data = pd.DataFrame(result,
                                             columns=["image", "xcentroid", "ycentroid", "mag", "merr", "flux", "ferr",
                                                      "SNR"]).set_index("image")

                    phot_data = phot_data.astype(float)

                    if extract is not None:
                        extracted_headers = self.__extract(extract)
                        if len(extracted_headers) != 0:
                            return pd.merge(phot_data, extracted_headers, left_index=True, right_index=True)

                    return phot_data



import numpy as np

from irony import Calibration, Fits, FitsArray


def test_calibration():
    fa = FitsArray.from_paths(["test/files/test1.fits"])
    bias = Fits.from_path("test/files/bias.fits")
    dark = Fits.from_path("test/files/dark.fits")
    flat = Fits.from_path("test/files/flat.fits")
    bdf_corrected = Fits.from_path("test/files/bdf_test.fits")

    ca = Calibration(fa)
    calibrated = ca.calibrate(zero=bias, dark=dark, flat=flat)

    np.testing.assert_equal(
        calibrated[0].data, bdf_corrected.data
    )



import pytest


from irony import APhot, FitsArray, Fixer

FILES = "test/files/test1.fits"
SOURCES = Fixer.list_to_source([[1495.226807, 1398.818519]])
APE = 10
ANN = 15
DAN = 20
EXPECTED_IRAF = [11.556, 0.005]
EXPECTED_SEP = [13.52960, 0.0]
EXPECTED_PHU = [13.58870, 4.470867]


def test_sep():
    fa = FitsArray.from_pattern(FILES)
    aphot = APhot(fa)
    phot = aphot.sep(SOURCES, APE)
    for calc, obs in zip(phot[["mag", "merr"]].to_numpy().tolist()[0], EXPECTED_SEP):
        assert calc == pytest.approx(obs)


def test_photutils():
    fa = FitsArray.from_pattern(FILES)
    aphot = APhot(fa)

    phot = aphot.photutils(SOURCES, APE, ANN)
    for calc, obs in zip(phot[["mag", "merr"]].to_numpy().tolist()[0], EXPECTED_PHU):
        assert calc == pytest.approx(obs)


def test_iraf():
    fa = FitsArray.from_pattern(FILES)
    aphot = APhot(fa)

    phot = aphot.iraf(SOURCES, APE, ANN, DAN)
    for calc, obs in zip(phot["mag"].to_numpy().tolist(), EXPECTED_IRAF):
        assert calc == pytest.approx(obs)



from subprocess import PIPE

from pyraf import iraf

from .base_logger import logger
from .errors import ImageCountError, NothingToDoError
from .fits import Fits, FitsArray
from .utils import Fixer


class Calibration:
    def __init__(self, fits_array: FitsArray) -> None:
        """
        Constructor method.
        Creates a Calibration Object.

        Parameters
        ----------
        fits_array: FitsArray
            A FitsArray.
        """
        logger.info(f"Creating an instance from {self.__class__.__name__}")
        if len(fits_array) < 1:
            logger.error("There is no image to process")
            raise ImageCountError("There is no image to process")

        self.fits_array = fits_array
        iraf.noao(Stdout=PIPE)
        iraf.imred(Stdout=PIPE)
        iraf.ccdred(Stdout=PIPE)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(@: {id(self)}, data: {self.fits_array})"

    def __repr__(self) -> str:
        return self.__str__()

    def calibrate(self, zero: Fits = None, dark: Fits = None, flat: Fits = None, output: str = None) -> FitsArray:
        """
        Returns the calibrated FitsArray.

        Parameters
        ----------
        zero: Fits
            Fits object of master zero. If None, zero correction will be skipped.
        dark: Fits
            Fits object of master dark. If None, dark correction will be skipped.
        flat: Fits
            Fits object of master flat. If None, flat correction will be skipped.
        output: str, optional
            Path of the new fits file.

        Returns
        -------
        FitsArray
            Calibrated FitsArray.
        """
        logger.info(f"Calibration started. Parameters: {output=}, {zero=}, {dark=}, {flat=}")
        if all([v is None for v in [zero, dark, flat]]):
            logger.error("Nothing neither of zero, dark ot flat ise provided. Nothing to do.")
            raise NothingToDoError("Nothing neither of zero, dark ot flat ise provided. Nothing to do.")

        zero_path = "" if zero is None else abs(zero)
        dark_path = "" if dark is None else abs(dark)
        flat_path = "" if flat is None else abs(flat)

        with self.fits_array.at_file() as at_file:
            with Fixer.to_new_directory(output, self.fits_array) as new_files:
                iraf.noao.imred.ccdred.ccdproc.unlearn()

                iraf.noao.imred.ccdred.ccdproc(f"'@{at_file}'", output=f"'@{new_files}'", noproc="no", ccdtype="",
                                               fixpix="no", oversca="no", trim="no",
                                               zerocor=Fixer.yesnoify(zero is not None), zero=zero_path,
                                               darkcor=Fixer.yesnoify(dark is not None), dark=dark_path,
                                               flatcor=Fixer.yesnoify(flat is not None), flat=flat_path)
                with open(new_files, "r") as to_save:
                    return FitsArray.from_paths(to_save.read().split())



from setuptools import setup, find_packages

with open("README.md", "r") as f2r:
    lng_desc = f2r.read()

setup(
    name='irony',
    version='0.1.1',
    license='GPL',
    long_description=lng_desc,
    long_description_content_type='text/markdown',
    author="Mohammad S.Niaei",
    author_email='m.shemuni@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/mshemuni/irony',
    download_url='https://github.com/user/reponame/archive/v_01.tar.gz',
    keywords=['IRAF', 'Photometry', 'Python'],
    install_requires=[
        'astroalign',
        'astropy',
        'matplotlib',
        'mpl_point_clicker',
        'numpy',
        'pandas',
        'photutils',
        'pyraf',
        'sep',
    ],
    classifiers=[
        'Natural Language :: English',
        'Intended Audience :: Developers',
        # 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Operating System :: POSIX :: Linux',
    ]

)



import pytest

from glob import glob
from pathlib import Path

import numpy as np

from src.irony import FitsArray
from irony import ImageCountError

FILES = "test/files/test*.fits"


def test_abs():
    fa = FitsArray.from_pattern(FILES)
    files = [str(Path(each).absolute()) for each in glob(FILES)]

    assert abs(fa) == files

    with pytest.raises(ImageCountError):
        _ = FitsArray.from_pattern("not_available/test*.fit")


def test_at_file():
    fa = FitsArray.from_pattern(FILES)

    with fa.at_file() as at_file:
        with open(at_file, "r") as f2r:
            assert f2r.read().split() == abs(fa)


def test_imstat():
    fa = FitsArray.from_pattern(FILES)
    stats = fa.imstat
    for i in range(len(fa)):
        assert stats.iloc[i].to_dict() == fa[i].imstat


def test_header():
    fa = FitsArray.from_pattern(FILES)
    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header

def test_hedit():
    fa = FitsArray.from_pattern(FILES)

    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header

    fa.hedit("IRON", "TEST")
    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header

    fa.hedit("IRON", delete=True)
    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header

    fa.hedit("IRON", "DATE-OBS", value_is_key=True)
    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header

    fa.hedit("IRON", delete=True)
    header = fa.header
    for i in range(len(fa)):
        assert header.iloc[i].to_dict() == fa[i].header


def test_hselect():
    fa = FitsArray.from_pattern(FILES)

    hselect = fa.hselect("DATE-OBS")
    assert hselect.to_numpy().flatten().tolist() == [fa[0].header["DATE-OBS"], fa[1].header["DATE-OBS"]]


def test_imarith():
    fa = FitsArray.from_pattern(FILES)

    new_fa = fa.imarith(10, "*")

    np.testing.assert_equal(
        new_fa[0].data, fa[0].data * 10
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data * 10
    )

    new_fa = fa.imarith(2, "/")
    np.testing.assert_equal(
        new_fa[0].data, fa[0].data / 2
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data / 2
    )

    new_fa = fa.imarith(fa, "+")
    np.testing.assert_equal(
        new_fa[0].data, fa[0].data * 2
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data * 2
    )

    new_fa = fa.imarith(fa, "-")
    np.testing.assert_equal(
        new_fa[0].data, fa[0].data * 0
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data * 0
    )

    new_fa = fa.imarith(fa[0], "-")
    np.testing.assert_equal(
        new_fa[0].data, fa[0].data * 0
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data - fa[0].data
    )

    new_fa = fa.imarith([2, 3], "*")
    np.testing.assert_equal(
        new_fa[0].data, fa[0].data * 2
    )

    np.testing.assert_equal(
        new_fa[1].data, fa[1].data * 3
    )


class EmissionValueError(Exception):
    """
    Raised on Emission error
    """
    pass


class NoiseValueError(Exception):
    """
    Raised on Noise error
    """
    pass


class ImageCountError(Exception):
    """
    Raised when number of fits files is insufficient
    """
    pass


class OperationValueError(Exception):
    """
    Raised on Operation error
    """
    pass


class RejectionValueError(Exception):
    """
    Raised on Rejection error
    """
    pass


class OperandValueError(Exception):
    """
    Raised on Operand error
    """
    pass


class ScaleValueError(Exception):
    """
    Raised on Scale error
    """
    pass


class NothingToDoError(Exception):
    """
    Raised on when no action can be taken
    """
    pass


class NumberOfElementError(Exception):
    """
    Raised when number of elements is insufficient
    """
    pass


class AlignError(Exception):
    """
    Raised on Align error
    """
    pass



import numpy as np

from irony import Combine, FitsArray


def test_combine():
    fa = FitsArray.from_pattern("test/files/test*.fits")

    c = Combine(fa)

    median = c.combine("median")
    np.testing.assert_equal(
        np.median([fa[0].data, fa[1].data], axis=0), median.data
    )

    average = c.combine("average")
    np.testing.assert_equal(
        np.mean([fa[0].data, fa[1].data], axis=0), average.data
    )

    the_sum = c.imsum()
    # Interesting. IRAF does modulo operation on imsum? -_-
    np.testing.assert_equal(
        np.sum([fa[0].data, fa[1].data], axis=0) % 65536, the_sum.data
    )
